@@ -38,11 +38,14 @@ import stat
 import sys
 import tempfile
 import time
+import unicodedata
+
+from concurrent.futures import ProcessPoolExecutor, wait, as_completed
 
 
 DEFAULT_CHUNK_SIZE = 16384  # block size in HFS+; 4X the block size in ext4
 DOT_THRESHOLD = 200
-VERSION = (0, 9, 2)
+VERSION = (1, 0, 0)
 IGNORED_FILE_SYSTEM_ERRORS = {errno.ENOENT, errno.EACCES}
 FSENCODING = sys.getfilesystemencoding()
 
@@ -52,6 +55,14 @@ if sys.version[0] == '2':
     # use `bytes` for bytestrings
 
 
+def normalize_path(path):
+    path_uni = path.decode(FSENCODING)
+    if FSENCODING in ('utf-8', 'UTF-8'):
+        return unicodedata.normalize('NFKD', path_uni)
+
+    return path_uni
+
+
 def sha1(path, chunk_size):
     digest = hashlib.sha1()
     with open(path, 'rb') as f:
@@ -95,16 +106,17 @@ def get_sqlite3_cursor(path, copy=False):
 
 
 def list_existing_paths(directory, expected=(), ignored=(), follow_links=False):
-    """list_existing_paths('/dir') -> ([path1, path2, ...], total_size)
+    """list_existing_paths(b'/dir') -> ([path1, path2, ...], total_size)
 
-    Returns a tuple with a list with existing files in `directory` and their
-    `total_size`.
+    Returns a tuple with a set of existing files in `directory` and its subdirectories
+    and their `total_size`. If directory was a bytes object, so will be the returned
+    paths.
 
     Doesn't add entries listed in `ignored`.  Doesn't add symlinks if
     `follow_links` is False (the default).  All entries present in `expected`
     must be files (can't be directories or symlinks).
     """
-    paths = []
+    paths = set()
     total_size = 0
     for path, _, files in os.walk(directory):
         for f in files:
@@ -129,12 +141,48 @@ def list_existing_paths(directory, expected=(), ignored=(), follow_links=False):
             else:
                 if not stat.S_ISREG(st.st_mode) or p in ignored:
                     continue
-                paths.append(p)
+                paths.add(p)
                 total_size += st.st_size
-    paths.sort()
     return paths, total_size
 
 
+def compute_one(path, chunk_size):
+    """Return a tuple with (unicode path, size, mtime, sha1). Takes a binary path."""
+    p_uni = normalize_path(path)
+    try:
+        st = os.stat(path)
+    except OSError as ex:
+        if ex.errno in IGNORED_FILE_SYSTEM_ERRORS:
+            # The file disappeared between listing existing paths and
+            # this run or is (temporarily?) locked with different
+            # permissions. We'll just skip it for now.
+            print(
+                '\rwarning: `{}` is currently unavailable for '
+                'reading: {}'.format(
+                    p_uni, ex,
+                ),
+                file=sys.stderr,
+            )
+            raise BitrotException
+
+        raise   # Not expected? https://github.com/ambv/bitrot/issues/
+
+    new_mtime = int(st.st_mtime)
+
+    try:
+        new_sha1 = sha1(path, chunk_size)
+    except (IOError, OSError) as e:
+        print(
+            '\rwarning: cannot compute hash of {} [{}]'.format(
+                p_uni, errno.errorcode[e.args[0]],
+            ),
+            file=sys.stderr,
+        )
+        raise BitrotException
+
+    return p_uni, st.st_size, int(st.st_mtime), new_sha1
+
+
 class BitrotException(Exception):
     pass
 
@@ -142,7 +190,7 @@ class BitrotException(Exception):
 class Bitrot(object):
     def __init__(
         self, verbosity=1, test=False, follow_links=False, commit_interval=300,
-        chunk_size=DEFAULT_CHUNK_SIZE,
+        chunk_size=DEFAULT_CHUNK_SIZE, workers=os.cpu_count(),
     ):
         self.verbosity = verbosity
         self.test = test
@@ -151,6 +199,7 @@ class Bitrot(object):
         self.chunk_size = chunk_size
         self._last_reported_size = ''
         self._last_commit_ts = 0
+        self.pool = ProcessPoolExecutor(max_workers=workers)
 
     def maybe_commit(self, conn):
         if time.time() < self._last_commit_ts + self.commit_interval:
@@ -180,67 +229,55 @@ class Bitrot(object):
         errors = []
         current_size = 0
         missing_paths = self.select_all_paths(cur)
+        hashes = self.select_all_hashes(cur)
         paths, total_size = list_existing_paths(
             b'.', expected=missing_paths, ignored={bitrot_db, bitrot_sha512},
             follow_links=self.follow_links,
         )
+        paths_uni = set(normalize_path(p) for p in paths)
+        futures = [self.pool.submit(compute_one, p, self.chunk_size) for p in paths]
 
-        for p in paths:
-            p_uni = p.decode(FSENCODING)
+        for future in as_completed(futures):
             try:
-                st = os.stat(p)
-            except OSError as ex:
-                if ex.errno in IGNORED_FILE_SYSTEM_ERRORS:
-                    # The file disappeared between listing existing paths and
-                    # this run or is (temporarily?) locked with different
-                    # permissions. We'll just skip it for now.
-                    print(
-                        '\rwarning: `{}` is currently unavailable for '
-                        'reading: {}'.format(
-                            p_uni, ex,
-                        ),
-                        file=sys.stderr,
-                    )
-                    continue
+                p_uni, new_size, new_mtime, new_sha1 = future.result()
+            except BitrotException:
+                continue
 
-                raise   # Not expected? https://github.com/ambv/bitrot/issues/
-
-            new_mtime = int(st.st_mtime)
-            current_size += st.st_size
+            current_size += new_size
             if self.verbosity:
                 self.report_progress(current_size, total_size)
 
-            missing_paths.discard(p_uni)
-            try:
-                new_sha1 = sha1(p, self.chunk_size)
-            except (IOError, OSError) as e:
-                print(
-                    '\rwarning: cannot compute hash of {} [{}]'.format(
-                        p, errno.errorcode[e.args[0]],
-                    ),
-                    file=sys.stderr,
-                )
-                continue
-
-            cur.execute('SELECT mtime, hash, timestamp FROM bitrot WHERE '
-                        'path=?', (p_uni,))
-            row = cur.fetchone()
-            if not row:
+            if p_uni not in missing_paths:
+                # We are not expecting this path, it wasn't in the database yet.
+                # It's either new or a rename. Let's handle that.
                 stored_path = self.handle_unknown_path(
-                    cur, p_uni, new_mtime, new_sha1,
+                    cur, p_uni, new_mtime, new_sha1, paths_uni, hashes
                 )
                 self.maybe_commit(conn)
-
                 if p_uni == stored_path:
-                    new_paths.append(p)   # FIXME: shouldn't that be p_uni?
+                    new_paths.append(p_uni)
+                    missing_paths.discard(p_uni)
                 else:
                     renamed_paths.append((stored_path, p_uni))
                     missing_paths.discard(stored_path)
                 continue
 
+            # At this point we know we're seeing an expected file.
+            missing_paths.discard(p_uni)
+            cur.execute('SELECT mtime, hash, timestamp FROM bitrot WHERE path=?',
+                        (p_uni,))
+            row = cur.fetchone()
+            if not row:
+                print(
+                    '\rwarning: path disappeared from the database while running:',
+                    p_uni,
+                    file=sys.stderr,
+                )
+                continue
+
             stored_mtime, stored_sha1, stored_ts = row
             if int(stored_mtime) != new_mtime:
-                updated_paths.append(p)
+                updated_paths.append(p_uni)
                 cur.execute('UPDATE bitrot SET mtime=?, hash=?, timestamp=? '
                             'WHERE path=?',
                             (new_mtime, new_sha1, ts(), p_uni))
@@ -248,11 +285,11 @@ class Bitrot(object):
                 continue
 
             if stored_sha1 != new_sha1:
-                errors.append(p)
+                errors.append(p_uni)
                 print(
                     '\rerror: SHA1 mismatch for {}: expected {}, got {}.'
                     ' Last good hash checked on {}.'.format(
-                        p, stored_sha1, new_sha1, stored_ts
+                        p_uni, stored_sha1, new_sha1, stored_ts
                     ),
                     file=sys.stderr,
                 )
@@ -262,6 +299,9 @@ class Bitrot(object):
 
         conn.commit()
 
+        if not self.test:
+            cur.execute('vacuum')
+
         if self.verbosity:
             cur.execute('SELECT COUNT(path) FROM bitrot')
             all_count = cur.fetchone()[0]
@@ -283,6 +323,10 @@ class Bitrot(object):
             )
 
     def select_all_paths(self, cur):
+        """Return a set of all distinct paths in the bitrot database.
+
+        The paths are Unicode and are normalized if FSENCODING was UTF-8.
+        """
         result = set()
         cur.execute('SELECT path FROM bitrot')
         row = cur.fetchone()
@@ -291,6 +335,20 @@ class Bitrot(object):
             row = cur.fetchone()
         return result
 
+    def select_all_hashes(self, cur):
+        """Return a dict where keys are hashes and values are sets of paths.
+
+        The paths are Unicode and are normalized if FSENCODING was UTF-8.
+        """
+        result = {}
+        cur.execute('SELECT hash, path FROM bitrot')
+        row = cur.fetchone()
+        while row:
+            rhash, rpath = row
+            result.setdefault(rhash, set()).add(rpath)
+            row = cur.fetchone()
+        return result
+
     def report_progress(self, current_size, total_size):
         size_fmt = '\r{:>6.1%}'.format(current_size/(total_size or 1))
         if size_fmt == self._last_reported_size:
@@ -303,6 +361,7 @@ class Bitrot(object):
     def report_done(
         self, total_size, all_count, error_count, new_paths, updated_paths,
         renamed_paths, missing_paths):
+        """Print a report on what happened.  All paths should be Unicode here."""
         print('\rFinished. {:.2f} MiB of data read. {} errors found.'
             ''.format(total_size/1024/1024, error_count))
         if self.verbosity == 1:
@@ -319,21 +378,21 @@ class Bitrot(object):
                 print('{} entries new:'.format(len(new_paths)))
                 new_paths.sort()
                 for path in new_paths:
-                    print(' ', path.decode(FSENCODING))
+                    print(' ', path)
             if updated_paths:
                 print('{} entries updated:'.format(len(updated_paths)))
                 updated_paths.sort()
                 for path in updated_paths:
-                    print(' ', path.decode(FSENCODING))
+                    print(' ', path)
             if renamed_paths:
                 print('{} entries renamed:'.format(len(renamed_paths)))
                 renamed_paths.sort()
                 for path in renamed_paths:
                     print(
                         ' from',
-                        path[0].decode(FSENCODING),
+                        path[0],
                         'to',
-                        path[1].decode(FSENCODING),
+                        path[1],
                     )
             if missing_paths:
                 print('{} entries missing:'.format(len(missing_paths)))
@@ -345,37 +404,38 @@ class Bitrot(object):
         if self.test and self.verbosity:
             print('warning: database file not updated on disk (test mode).')
 
-    def handle_unknown_path(self, cur, new_path, new_mtime, new_sha1):
+    def handle_unknown_path(self, cur, new_path, new_mtime, new_sha1, paths_uni, hashes):
         """Either add a new entry to the database or update the existing entry
         on rename.
 
-        Returns `new_path` if the entry was indeed new or the `stored_path` (e.g.
-        outdated path) if there was a rename.
+        `cur` is the database cursor. `new_path` is the new Unicode path.
+        `paths_uni` are Unicode paths seen on disk during this run of Bitrot.
+        `hashes` is a dictionary selected from the database, keys are hashes, values
+        are sets of Unicode paths that are stored in the DB under the given hash.
+
+        Returns `new_path` if the entry was indeed new or the `old_path` (e.g.
+        outdated path stored in the database for this hash) if there was a rename.
         """
-        cur.execute('SELECT mtime, path, timestamp FROM bitrot WHERE hash=?',
-                    (new_sha1,))
-        rows = cur.fetchall()
-        for row in rows:
-            stored_mtime, stored_path, stored_ts = row
-            if os.path.exists(stored_path):
-                # file still exists, move on
-                continue
 
-            # update the path in the database
+        for old_path in hashes.get(new_sha1, ()):
+            if old_path not in paths_uni:
+                # File of the same hash used to exist but no longer does.
+                # Let's treat `new_path` as a renamed version of that `old_path`.
+                cur.execute(
+                    'UPDATE bitrot SET mtime=?, path=?, timestamp=? WHERE path=?',
+                    (new_mtime, new_path, ts(), old_path),
+                )
+                return old_path
+
+        else:
+            # Either we haven't found `new_sha1` at all in the database, or all
+            # currently stored paths for this hash still point to existing files.
+            # Let's insert a new entry for what appears to be a new file.
             cur.execute(
-                'UPDATE bitrot SET mtime=?, path=?, timestamp=? WHERE path=?',
-                (new_mtime, new_path, ts(), stored_path),
+                'INSERT INTO bitrot VALUES (?, ?, ?, ?)',
+                (new_path, new_mtime, new_sha1, ts()),
             )
-
-            return stored_path
-
-        # no rename, just a new file with the same hash
-        cur.execute(
-            'INSERT INTO bitrot VALUES (?, ?, ?, ?)',
-            (new_path, new_mtime, new_sha1, ts()),
-        )
-        return new_path
-
+            return new_path
 
 def get_path(directory=b'.', ext=b'db'):
     """Compose the path to the selected bitrot file."""
@@ -492,6 +552,9 @@ def run_from_command_line():
         '--commit-interval', type=float, default=300,
         help='min time in seconds between commits '
              '(0 commits on every operation)')
+    parser.add_argument(
+        '-w', '--workers', type=int, default=os.cpu_count(),
+        help='run this many workers (use -w1 for slow magnetic disks)')
     parser.add_argument(
         '--chunk-size', type=int, default=DEFAULT_CHUNK_SIZE,
         help='read files this many bytes at a time')
@@ -517,6 +580,7 @@ def run_from_command_line():
             follow_links=args.follow_links,
             commit_interval=args.commit_interval,
             chunk_size=args.chunk_size,
+            workers=args.workers,
         )
         if args.fsencoding:
             FSENCODING = args.fsencoding
